__all__ = ["EllipticCurveNISTP256", "ECPoint", "ECDH", "ECKey", "ECSchnorr", "ECIES"]
from .ECPoint import ECPoint
from .ECDH import ECDH
from .ECKey import ECKey
from .ECSchnorr import ECSchnorr
from .ECIES import ECIES
from .EllipticCurveNISTP256 import EllipticCurveNISTP256



from __future__ import annotations
# Fiat–Shamir Zero Knowledge Proof
from AsymmetricEncryptions.General import PrimeNumberGen, BytesAndInts
import secrets
import hashlib

class FiatShamirZeroKnowledgeProof:
    """
    Example:
    ```
    from AsymmetricEncryptions.Protocols import FiatShamirZeroKnowledgeProof
    m = b'test'
    fs = FiatShamirZeroKnowledgeProof.new()
    y = fs.AliceStage1(m)
    # send y to bob
    v, t = fs.AliceStage2()
    # keep v send t
    c = fs.BobStage1()
    # Bob sends c as a challenge
    r = fs.AliceStage3(v, c, m)
    # Alice sends r to bob
    ver = fs.BobStage2(y, r, c, t)
    print(ver)
    ```
    """
    def __init__(self, g: int, n: int):
        self.g: int = g
        self.n: int = n

    @staticmethod
    def new() -> FiatShamirZeroKnowledgeProof:
        """
        :return: Generates a FiatShamirZeroKnowledgeProof object
        """
        n: int = PrimeNumberGen.generate(1024)
        g: int = secrets.randbelow(n)
        return FiatShamirZeroKnowledgeProof(g, n)

    def AliceStage1(self, msg: bytes) -> int:
        """
        :param msg: The message to be proven
        :return: y, the other person needs to keep it
        """
        x: int = BytesAndInts.byte2Int(hashlib.sha256(msg).digest()) % self.n
        y: int = pow(self.g, x, self.n)
        return y
        # Bob keeps y

    def AliceStage2(self) -> int and int:
        v: int = secrets.randbelow(self.n)
        t: int = pow(self.g, v, self.n)
        return v, t
        # Alice keeps v
        # Bob keeps t

    def BobStage1(self) -> int:
        c: int = secrets.randbelow(self.n)
        return c

    def AliceStage3(self, v: int, c: int, msg: bytes) -> int:
        x: int = BytesAndInts.byte2Int(hashlib.sha256(msg).digest()) % self.n
        r: int = (v - (c * x))
        return r

    def BobStage2(self, y: int, r: int, c: int,t: int) -> bool:
        x: int = pow(self.g, r, self.n) * pow(y, c, self.n)
        x %= self.n
        return x == t





from __future__ import annotations
from AsymmetricEncryptions.General import XOR
from .KDF import KDF
import warnings

class ThreePassProtocol:

    """without any additional authentication the protocol is susceptible to a man-in-the-middle attack if the opponent has the ability to create false messages, or to intercept and replace the genuine transmitted messages."""

    def __init__(self, key: bytes, *, warningsBool=True) -> None:
        self.wb = warningsBool
        if warningsBool: warnings.warn("WARNING: Three pass is considered un-safe without authentication!")
        self.key: bytes = KDF.derive_key(key)

    def stage1(self, msg: bytes, *, encryption_function=XOR.repeated_key_xor) -> bytes:
        """
        Alice encrypts her message.
        :param msg: The message.
        :param encryption_function: the symmetric encryption function used
        :return: bytes, send to Bob
        """
        if self.wb: warnings.warn("WARNING: Three pass is considered un-safe without authentication!")
        return encryption_function(msg, self.key)

    @staticmethod
    def stage2(key: bytes, ciphertext: bytes, *, encryption_function=XOR.repeated_key_xor) -> bytes:
        """
        Bob gets a cipher text, and hopefully authentication as well
        :param key: Bob's key
        :param ciphertext: The ciphertext that Alice gave me.
        :param encryption_function: the symmetric encryption function used (can be different from Alice's)
        :return: send ciphertext and auth as well
        """
        key: bytes = KDF.derive_key(key)
        return encryption_function(ciphertext, key)

    def stage3(self, ciphertext, *, decryption_function=XOR.repeated_key_xor) -> bytes:
        """
        Alice hopefully gets an authentication and a ciphertext
        :param ciphertext: The ciphertext Bob sent.
        :param decryption_function: The symmetric decryption function.
        :return: ciphertext and sign it as well.
        """
        if self.wb: warnings.warn("WARNING: Three pass is considered un-safe without authentication!")
        return decryption_function(ciphertext, self.key)

    @staticmethod
    def stage4(ciphertext: bytes, key: bytes,*, decryption_function=XOR.repeated_key_xor) -> bytes:
        """
        Bob hopefully gets an authentication and a ciphertext
        :param ciphertext: The ciphertext Alice sent.
        :param key: Bob's key
        :param decryption_function: The symmetric decryption function.
        :return: The message
        """
        key: bytes = KDF.derive_key(key)
        return decryption_function(ciphertext, key)







from __future__ import annotations
from .EllipticCurveNISTP256 import EllipticCurveNISTP256
import hashlib

class ECPoint:
    """A point on the ECC NIST-P-256 curve"""
    # https://github.com/cgossi/fundamental_cryptography_with_python/blob/main/implementing_p_256_ecdhe.py
    def __init__(self, curve: EllipticCurveNISTP256, x: int, y: int) -> None:
        self.curve: EllipticCurveNISTP256 = curve
        self.x: int = x
        self.y: int = y

    def copy(self) -> ECPoint:
        return ECPoint(self.curve, self.x, self.y)

    def is_inf(self) -> bool:
        return self == self.curve.infinity()

    def __str__(self):
        return f"({self.x}, {self.y})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, ECPoint): return False
        return self.x == other.x and self.y == other.y

    # Point multiplication
    def __mul__(self, s):
        bits = [s & (1 << i) for i in range(s.bit_length() - 1, -1, -1)]
        res = self.curve.infinity()
        for bit in bits:
            res = res + res
            if bit:
                res = res + self
        return res

    def __bytes__(self) -> bytes:
        return hashlib.sha256(str(self).encode()).digest()

    def __imul__(self, other):
        return self.__mul__(other)

    # Point addition
    def __add__(self, other):
        if self.is_inf():
            return other.copy()
        if other.is_inf():
            return self.copy()
        x1, y1 = self.x, self.y
        x2, y2 = other.x, other.y
        p = self.curve.p
        if x1 % p == x2 % p and y1 % p == (-y2) % p:
            return self.curve.infinity()
        if self != other:
            s = (y2 - y1) * pow(x2 - x1, -1, p) % p
        else:
            # Point doubling
            s = (3 * pow(x1, 2) + self.curve.a) * pow(2 * y1, -1, p) % p
        x3 = (pow(s, 2) - x1 - x2) % p
        y3 = (s * (x1 - x3) - y1) % p
        return ECPoint(curve=self.curve, x=x3, y=y3)



from __future__ import annotations
from . import ECPoint, ECKey
from .EllipticCurveNISTP256 import EllipticCurveNISTP256
from AsymmetricEncryptions.General import BytesAndInts
import secrets
from hashlib import sha256

class ECSchnorr:
    """Schnorr's signature scheme implemented with ECC."""
    def __init__(self, key: ECKey):
        self.key: ECKey = key

    def sign(self, msg: bytes) -> tuple[int, ECPoint]:
        """
        Signs a message
        :param msg: the message
        :return: signature
        """
        G: ECPoint = EllipticCurveNISTP256().g()
        r: int = secrets.randbelow(EllipticCurveNISTP256.p)
        R: ECPoint = G * r
        c: int = BytesAndInts.byte2Int(sha256(bytes(R) + msg).digest())
        s: int = (c * self.key.private_key) + r
        return s, R

    @staticmethod
    def verify(signature: tuple[int, ECPoint], msg: bytes, pubkey: ECPoint) -> bool:
        """
        Verifies a signature.
        :param signature: The signature.
        :param msg: The message.
        :param pubkey: the public key.
        :return: True if it passes, else: False.
        """
        G: ECPoint = EllipticCurveNISTP256().g()
        s, R = signature
        c: int = BytesAndInts.byte2Int(sha256(bytes(R) + msg).digest())
        return G * s == (pubkey * c) + R

if __name__ == '__main__':
    key = ECKey.new()
    signer = ECSchnorr(key)
    msg = b"test"
    signature = signer.sign(msg)
    verify = ECSchnorr.verify(signature, msg, key.public_key)
    print(verify)



from __future__ import annotations
from .ElGamalKey import ElGamalKey
from AsymmetricEncryptions.General import BytesAndInts
import secrets
from math import gcd
class ElGamal:
    """ElGamal encryption"""
    def __init__(self, key: ElGamalKey) -> None:
        """
        :param key: private / public key
        """
        self.key = key

    @staticmethod
    def generate_key_pair(nBit) -> ElGamalKey and ElGamalKey:
        """
        ElGamalKey.new returns a pair of priv and pub
        :param nBit: How big is p.
        :return: private_key, public_key
        """
        Priv: ElGamalKey = ElGamalKey.new(nBit)
        Pub: ElGamalKey = Priv.public
        return Priv, Pub

    def encrypt(self, msg: bytes, assertion=True) -> tuple[bytes, bytes]:
        """
        Encrypts a message
        :param msg: The plaintext
        :param assertion: If assertion is true then it will assert if the message is smaller than allowed. (e.g. m < p)
        :return: Ciphertext
        """
        m: bytes = BytesAndInts.byte2Int(msg)
        if assertion: assert m < self.key.p
        b: int = secrets.randbelow(self.key.p)
        c1: int = pow(self.key.g, b, self.key.p)
        c2: int = (m * pow(self.key.y, b, self.key.p)) % self.key.p
        return BytesAndInts.int2Byte(c1), BytesAndInts.int2Byte(c2)

    def decrypt(self, cipher: tuple[bytes, bytes]) -> bytes:
        """
        Decrypts (Only possible if key is private)
        :param cipher: the ciphertext
        :return: message
        """
        assert self.key.x
        c1: int = BytesAndInts.byte2Int(cipher[0])
        c2: int = BytesAndInts.byte2Int(cipher[1])
        a: int = pow(c1, self.key.x, self.key.p)
        m = (c2 * pow(a, self.key.p - 2, self.key.p)) % self.key.p
        return BytesAndInts.int2Byte(m)


    def sign(self, msg: bytes, assertion=True) -> tuple[bytes, bytes, bytes]:
        """
        Signs a message
        :param msg: The plaintext
        :param assertion: if true then will check if m is allowed
        :return: signature
        """
        m: bytes = BytesAndInts.byte2Int(msg)
        if assertion: assert m < self.key.p
        k: int = secrets.randbelow(self.key.p)
        while gcd(k, self.key.p - 1) != 1: k = secrets.SystemRandom().randint(2, self.key.p - 2)
        s1: int = pow(self.key.g, k, self.key.p)
        phi_n: int = self.key.p - 1
        inv: int = pow(k, -1, phi_n)
        s2: int = (inv * (m - s1 * self.key.x)) % phi_n
        if s2 == 0: self.sign(msg, assertion)
        return BytesAndInts.int2Byte(s1), BytesAndInts.int2Byte(s2),  BytesAndInts.int2Byte(m)

    def verify(self, signature: tuple[bytes, bytes, bytes]) -> bool:
        """
        Verifies a signature
        :param signature: The signature
        :return: bool
        """
        s1: int = BytesAndInts.byte2Int(signature[0])
        s2: int = BytesAndInts.byte2Int(signature[1])
        m: int = BytesAndInts.byte2Int(signature[2])
        V = pow(self.key.y, s1, self.key.p) * pow(s1, s2, self.key.p)
        V = V % self.key.p
        W = pow(self.key.g, m, self.key.p)
        print(W)
        print(V)
        return V == W



__all__ = ["RSA", "DSA", "ElGamal", "ECC", "ECPoint", "ECKey", "ECSchnorr", "ECIES", "ECDH", "EllipticCurveNISTP256", "DLIES", "DLIESKey"]
from .ElGamal import ElGamalKey, ElGamal
from .RSA import RSAKey, RSA
from .DSA import DSA, DSAKey
from .DLIES import DLIESKey, DLIES
from .ECC import ECPoint, ECKey, ECSchnorr, ECIES, ECDH, EllipticCurveNISTP256



class XOR:
    @staticmethod
    def repeated_key_xor(plain_text: bytes, key: bytes) -> bytes:
        """
        Encrypts using OTP
        :param plain_text: The plain text
        :param key: The encryption/decryption key
        :return: ciphertext
        """
        pt = plain_text
        len_key = len(key)
        encoded = []

        for i in range(0, len(pt)):
            encoded.append(pt[i] ^ key[i % len_key])
        return bytes(encoded)



from __future__ import annotations
from math import gcd
from AsymmetricEncryptions.General import PrimeNumberGen, XOR, Exportation
from AsymmetricEncryptions.Exceptions import NeededValueIsNull
import hashlib

class RSAKey:
    def __init__(self, p: int or None = None, q: int or None = None, n: int = None, e: int = None, d: int or None = None, tot_n: int or None = None) -> None:
        if not e or not n:
            raise NeededValueIsNull("e or n needed to create key")
        self.p = p
        self.q = q
        self.n = n
        self.tot_n = tot_n
        self.e = e
        self.d = d
        self.has_private = False
        if d:
            self.has_private = True
            self.public = RSAKey(None, None, n, e, None, None)

    @staticmethod
    def new(bit_number: int) -> RSAKey:
        p: int = PrimeNumberGen.generate(bit_number)
        q: int = PrimeNumberGen.generate(bit_number)
        n: int = p * q
        tot_n: int = (p - 1) * (q - 1)
        e: int = PrimeNumberGen.generate(bit_number)
        while gcd(e, tot_n) != 1:
            e: int = PrimeNumberGen.generate(bit_number)
        d: int = pow(e, -1, tot_n)
        return RSAKey(p, q, n, e, d, tot_n)


    def export(self, file_name: str, pwd: bytes = b"\x00", *, enc_func=XOR.repeated_key_xor) -> None:
        data_dict: dict = {"p": self.p, "q": self.q, "n": self.n, "tot_n": self.tot_n, "e": self.e, "d": self.d}
        Exportation.Exportation.export(file_name=file_name, pwd=pwd, data_dict=data_dict, exportation_func=enc_func)

    @staticmethod
    def load(file_name: str, pwd: bytes = b"\x00", *, dec_func=XOR.repeated_key_xor) -> RSAKey:
        return RSAKey(**Exportation.Exportation.load(file_name=file_name, pwd=pwd, dec_func=dec_func))

    def __eq__(self, other: RSAKey) -> bool:
        if not isinstance(other, RSAKey): return False
        return hashlib.sha256(f"{self}".encode()).hexdigest() == hashlib.sha256(f"{other}".encode()).hexdigest()


    def __str__(self) -> str:
        r: str = ""
        if self.has_private:
            r += f"""
Private Key:

p = {self.p}
q = {self.q}
n = {self.n}
tot_n = {self.tot_n}
e = {self.e}
d = {self.d}

"""
        r += f"""
Public Key:

n = {self.n}
e = {self.e}
"""
        return r



from __future__ import annotations
import hashlib

class KDF:
    """Key Derivation Function"""
    def __init__(self):
        pass

    @staticmethod
    def derive_key(passphrase: bytes) -> bytes:
        """
        Derives a random (I think safe) key
        :param passphrase: The passphrase to derive the key from
        :return: bytes
        """
        count: int = 0
        cur: bytes = b'\x00'
        while b'\x00' in cur:
            cur = hashlib.scrypt(passphrase, salt=hashlib.sha256(passphrase + bytes(count)).digest(), r=16, n=2048, p=1)
            count += 1
        return cur






__all__ = ["SSS", "OAEP", "FSZKP", "FiatShamirZeroKnowledgeProof", "OT1O2", "ObliviousTransfer", "RSADH", "DiffieHellman", "ThreePass", "ThreePassProtocol", "KDF"]
from .SSS import SSS
from .OAEP import OAEP
from .FSZKP import FiatShamirZeroKnowledgeProof
from .OT1O2 import ObliviousTransfer
from .RSADH import DiffieHellman
from .KDF import KDF
from .ThreePass import ThreePassProtocol




class BytesAndInts:
    """A class to convert bytes and ints easily"""

    def __init__(self):
        pass

    @staticmethod
    def int2Byte(i: int) -> bytes:
        """
        Turns int into bytes using the to_bytes with byteorder=little
        :param i: the int to convert
        :return: bytes
        """
        return i.to_bytes(i.bit_length(), "little").rstrip(b'\x00')

    @staticmethod
    def byte2Int(b: bytes) -> int:
        """
        Turns bytes into int using the from_bytes with byteorder=little
        :param b: the byte to convert
        :return: int
        """
        return int.from_bytes(b, "little")



from __future__ import annotations
import json
import secrets
from AsymmetricEncryptions.General import PrimeNumberGen, XOR, Exportation
import hmac
import hashlib

class ElGamalKey:
    """ElGamal key object"""
    def __init__(self, p: int, g: int, y: int, x: int or None = None) -> None:
        """Use .new()"""
        self.p: int = p
        self.g: int = g
        self.y: int = y
        self.x: int = x
        self.has_private: bool = False
        if x:
            self.has_private = True
            self.public: ElGamalKey = ElGamalKey(p=p, g=g, y=y, x=None)

    def __eq__(self, other):
        if not isinstance(other, ElGamalKey): return False
        return hashlib.sha256(f"{self}".encode()).hexdigest() == hashlib.sha256(f"{other}".encode()).hexdigest()

    def __str__(self) -> str:
        r: str = ""
        if self.has_private:
            r += f"""
Private Key:

p = {self.p}
q = {self.g}
x = {self.x}
y = {self.y}
"""
        r += f"""
Public Key:

p = {self.p}
q = {self.g}
y = {self.y}
"""
        return r

    @staticmethod
    def new(nBit) -> ElGamalKey:
        """
        Returns a new ElGamalKey object
        :param nBit: key length
        :return: ElGamalKey
        """
        p: int = PrimeNumberGen.generate(nBit)
        g: int = secrets.randbelow(p)
        x: int = secrets.randbelow(p)
        y: int = pow(g, x, p)
        return ElGamalKey(p=p, g=g, y=y, x=x)


    def export(self, file_name: str, pwd: bytes = b"\x00", *, enc_func=XOR.repeated_key_xor) -> None:
        """
        Exports the key.
        :param file_name: File name
        :param pwd: Passphrase
        :param enc_func: Symmetric encryption function (Without touching it, it's XOR / OTP)
        :return: None
        """
        data_dict: dict = {"p": self.p, "g": self.g, "y": self.y, "x": self.x}
        Exportation.export(file_name=file_name, pwd=pwd, data_dict=data_dict, exportation_func=enc_func)

    @staticmethod
    def load(file_name: str, pwd: bytes = b"\x00", *, dec_func=XOR.repeated_key_xor) -> ElGamalKey:
        """
        Loads a key from a file.
        :param file_name: File name
        :param pwd: Passphrase
        :param dec_func: Symmetric decryption function (Without touching it, it's XOR / OTP)
        :return: ElGamalKey if it succeeds.
        """
        return ElGamalKey(**Exportation.load(file_name=file_name, pwd=pwd, dec_func=dec_func))



from __future__ import annotations
from AsymmetricEncryptions.PublicPrivateKey.DSA import DSAKey
from AsymmetricEncryptions.General import BytesAndInts
import secrets
import hashlib
class DSA:
    """Digital Signature Algorithm"""
    def __init__(self, key: DSAKey) -> None:
        self.key = key

    def sign(self, message: bytes) -> tuple[int, int]:
        """
        Signs a message.
        :param message: Message to sign
        :return: Signature.
        """
        assert self.key.has_private
        m: int = BytesAndInts.byte2Int(message)
        k: int = secrets.randbelow(self.key.q - 1)
        r: int = pow(self.key.g, k, self.key.p) % self.key.q
        if r == 0: self.sign(message)
        s: int = (pow(k, -1, self.key.q) * ((self.H(m) + self.key.x * r) % self.key.q)) % self.key.q
        if s == 0: self.sign(message)
        return r, s

    def verify(self, sig: tuple[int, int], message: bytes) -> None:
        """
        Verifies a signed message.
        :param sig: The signature.
        :param message: The original message
        :return: None, will throw an assertion error if it failed to verify.
        """
        r: int = sig[0]
        s: int = sig[1]
        assert 0 < r < self.key.q and 0 < s < self.key.q
        m: int = BytesAndInts.byte2Int(message)
        w: int = pow(s, -1, self.key.q)
        u1: int = (self.H(m) * w) % self.key.q
        u2: int = (r * w) % self.key.q
        v = ((pow(self.key.g, u1, self.key.p) * pow(self.key.y, u2, self.key.p)) % self.key.p) % self.key.q
        assert v == r

    @staticmethod
    def H(m: int) -> int:
        # H = sha256
        b: bytes = BytesAndInts.int2Byte(m)
        return BytesAndInts.byte2Int(hashlib.sha256(b, usedforsecurity=True).digest())


    @staticmethod
    def generate_key_pair(nBit: int = 1024, use_precalculated: bool = True) -> DSAKey and DSAKey:
        # DSAKey.new returns a pair of priv and pub
        Priv: DSAKey = DSAKey.new(nBit=nBit, use_precalculated=use_precalculated)
        Pub: DSAKey = Priv.public
        return Priv, Pub



from __future__ import annotations
from .EllipticCurveNISTP256 import EllipticCurveNISTP256
from .ECPoint import ECPoint
from AsymmetricEncryptions.General import XOR, Exportation
import secrets
from hashlib import sha256
class ECKey:
    """Key object of ECC"""
    def __init__(self, pub: ECPoint, priv: int = None) -> None:
        """Use .new()"""
        self.private_key = priv
        self.public_key = pub

    @staticmethod
    def new() -> ECKey:
        """Creates a new ECKey pair"""
        priv: int = secrets.randbelow(EllipticCurveNISTP256.p)
        pub: ECPoint = EllipticCurveNISTP256().g() * priv
        return ECKey(pub, priv)

    def export(self, file_name: str, pwd: bytes = b"\x00", *, enc_func=XOR.repeated_key_xor) -> None:
        """
        Exports the key into a file.
        :param file_name: The file to export into.
        :param pwd: Passphrase
        :param enc_func: symmetric encryption function (XOR / OTP here)
        :return: None
        """
        data_dict: dict = {"pub_x": self.public_key.x, "pub_y": self.public_key.y, "priv": self.private_key}
        Exportation.export(file_name=file_name, pwd=pwd, data_dict=data_dict, exportation_func=enc_func)

    @staticmethod
    def load(file_name: str, pwd: bytes = b"\x00", *, dec_func=XOR.repeated_key_xor) -> ECKey:
        """
        Loads a key from a file
        :param file_name: The file name
        :param pwd: The password
        :param dec_func: symmetric decryption function (XOR / OTP here)
        :return: ECKey
        """
        data: dict = Exportation.load(file_name=file_name, pwd=pwd, dec_func=dec_func)
        pub: ECPoint = ECPoint(EllipticCurveNISTP256, data["pub_x"], data["pub_y"])
        return ECKey(pub, data["priv"])


    def __str__(self):
        return f"{self.private_key = }\nself.public_key = {self.public_key}"

    def __eq__(self, other: ECKey) -> bool:
        if not isinstance(other, ECKey): return False
        return sha256(f"{self}".encode()).hexdigest() == sha256(f"{other}".encode()).hexdigest()



# https://www.geeksforgeeks.org/how-to-generate-large-prime-numbers-for-rsa-algorithm/
# Large Prime Generation for RSA
import secrets

class PrimeNumberGen:
    """Generation of Prime numbers and checking if prime"""
    # Pre generated primes
    first_primes_list = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29,
                         31, 37, 41, 43, 47, 53, 59, 61, 67,
                         71, 73, 79, 83, 89, 97, 101, 103,
                         107, 109, 113, 127, 131, 137, 139,
                         149, 151, 157, 163, 167, 173, 179,
                         181, 191, 193, 197, 199, 211, 223,
                         227, 229, 233, 239, 241, 251, 257,
                         263, 269, 271, 277, 281, 283, 293,
                         307, 311, 313, 317, 331, 337, 347, 349]

    @staticmethod
    def nBitRandom(n):
        return secrets.SystemRandom().randrange(pow(2, (n - 1)) + 1, pow(2, n) - 1)

    @staticmethod
    def getLowLevelPrime(n):
        """Generate a prime candidate divisible
        by first primes"""
        while True:
            # Obtain a random number
            pc = PrimeNumberGen.nBitRandom(n)

            # Test divisibility by pre-generated
            # primes
            for divisor in PrimeNumberGen.first_primes_list:
                if pc % divisor == 0 and divisor ** 2 <= pc:
                    break
            else:
                return pc

    @staticmethod
    def isMillerRabinPassed(mrc):
        """Run 20 iterations of Rabin Miller Primality test. Use to check if mrc is prime"""
        maxDivisionsByTwo = 0
        ec = mrc - 1
        while ec % 2 == 0:
            ec >>= 1
            maxDivisionsByTwo += 1
        assert (2 ** maxDivisionsByTwo * ec == mrc - 1)

        def trialComposite(round_tester):
            if pow(round_tester, ec, mrc) == 1:
                return False
            for i in range(maxDivisionsByTwo):
                if pow(round_tester, 2 ** i * ec, mrc) == mrc - 1:
                    return False
            return True

        # Set number of trials here
        numberOfRabinTrials = 20
        for i in range(numberOfRabinTrials):
            round_tester = secrets.SystemRandom().randrange(2, mrc)
            if trialComposite(round_tester):
                return False
        return True

    @staticmethod
    def generate(nBit: int) -> int:
        """
        Generates an nBit prime number
        :param nBit: The bit size of the number
        :return: Random nBit prime number as an int
        """
        while True:
            prime_candidate = PrimeNumberGen.getLowLevelPrime(nBit)
            if not PrimeNumberGen.isMillerRabinPassed(prime_candidate):
                continue
            return prime_candidate



from __future__ import annotations
from AsymmetricEncryptions.PublicPrivateKey.RSA import RSA, RSAKey
from AsymmetricEncryptions.General import BytesAndInts
import secrets
# 1 out of 2 Oblivious Transfer https://en.wikipedia.org/wiki/Oblivious_transfer


class ObliviousTransfer:
    """
    Oblivious Transfer.
    The methods with static method on them are ment to be used by the other person.
    (In the example below Alice wants to send Bob only one of the messages)
    Example:
    ```
    # Alice
    otProt = ObliviousTransfer(b"test A", b"test B")
    sendBob = otProt.Stage1and2and3()
    # Bob
    b = int(input("Choice 0 or 1: ")) % 2
    AlicePubKey = sendBob[0]
    sendAlice, keepPrivate = ObliviousTransfer.Stage4and5(sendBob, b)
    # Alice
    sendBob = otProt.Stage6and7(sendAlice)
    # Bob
    m = ObliviousTransfer.Stage8(sendBob, keepPrivate, b, AlicePubKey)
    print(m)
    ```
    """
    def __init__(self, m0: bytes, m1: bytes) -> None:
        # stage one
        self.m0: int = BytesAndInts.byte2Int(m0)
        self.m1: int = BytesAndInts.byte2Int(m1)
        self.priv = RSA.generate_key_pair(1024)
        self.pub: RSAKey = self.priv[1]
        self.priv: RSAKey = self.priv[0]
        self.x0: int = None
        self.x1: int = None

    def Stage1and2and3(self) -> tuple[RSAKey, int, int]:
        pub: RSAKey = self.priv.public
        x0: int = secrets.randbelow(pub.n)
        x1: int = secrets.randbelow(pub.n)
        self.x0 = x0
        self.x1 = x1
        return pub, x0, x1
        # send pub, x0, x1 to the other person

    @staticmethod
    def Stage4and5(pubX0X1: tuple[RSAKey, int, int], b: int) -> int and int:
        pub: RSAKey = pubX0X1[0]
        x0: int = pubX0X1[1]
        x1: int = pubX0X1[2]
        assert b in [0, 1]
        xs: tuple[int, int] = (x0, x1)
        k: int = secrets.randbelow(pub.n)
        v: int = (xs[b] + pow(k, pub.e, pub.n)) % pub.n
        return v, k
        # send only v! keep k private

    def Stage6and7(self, v: int) -> tuple[int, int]:
        assert self.x0 and self.x1
        k0: int = pow((v - self.x0), self.priv.d, self.priv.n)
        k1: int = pow((v - self.x1), self.priv.d, self.priv.n)
        mp0: int = (self.m0 + k0) % self.priv.n
        mp1: int = (self.m1 + k1) % self.priv.n
        return mp0, mp1
        # send both

    @staticmethod
    def Stage8(mps: tuple[int, int], k: int, b: int, pub: RSAKey) -> bytes:
        mb: int = (mps[b] - k) % pub.n
        return BytesAndInts.int2Byte(mb)


if __name__ == '__main__':
    # Alice
    otProt = ObliviousTransfer(b"test A", b"test B")
    sendBob = otProt.Stage1and2and3()
    # Bob
    b = int(input("Choice 0 or 1: ")) % 2
    AlicePubKey = sendBob[0]
    sendAlice, keepPrivate = ObliviousTransfer.Stage4and5(sendBob, b)
    # Alice
    sendBob = otProt.Stage6and7(sendAlice)
    # Bob
    m = ObliviousTransfer.Stage8(sendBob, keepPrivate, b, AlicePubKey)
    print(m)



from __future__ import annotations
from AsymmetricEncryptions.PublicPrivateKey.ECC import ECPoint


class EllipticCurveNISTP256:
    """

    """
    p: int = 0xffffffff00000001000000000000000000000000ffffffffffffffffffffffff
    n: int = 0xffffffff00000000ffffffffffffffffbce6faada7179e84f3b9cac2fc632551
    a: int = 0xffffffff00000001000000000000000000000000fffffffffffffffffffffffc
    b: int = 0x5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b
    g_x: int = 0x6b17d1f2e12c4247f8bce6e563a440f277037d812deb33a0f4a13945d898c296
    g_y: int = 0x4fe342e2fe1a7f9b8ee7eb4a7c0f9e162bce33576b315ececbb6406837bf51f5

    def g(self) -> ECPoint.ECPoint:
        return ECPoint.ECPoint(curve=self, x=EllipticCurveNISTP256.g_x, y=EllipticCurveNISTP256.g_x)

    def infinity(self) -> ECPoint.ECPoint:
        return ECPoint.ECPoint(curve=self, x=None, y=None)

    def __init__(self):
        self.p = EllipticCurveNISTP256.p
        self.n = EllipticCurveNISTP256.n
        self.a = EllipticCurveNISTP256.a
        self.b = EllipticCurveNISTP256.b
        self.G = self.g()

    def __eq__(self, other) -> bool:
        return isinstance(other, EllipticCurveNISTP256)

    def f(self, x: int):
        return (pow(x, 3, self.p) + self.a * x + self.b) % self.p




import hashlib
import secrets
from AsymmetricEncryptions.General import XOR
class OAEP:
    """
    Optimal Asymmetric Encryption Padding.
    Using Sha256 as the hash algorithm.
    """
    @staticmethod
    def oaep_pad(message: bytes) -> bytes:
        """
        Pads the message.
        :param message: The message to be padded.
        :return: Padded message
        """
        nonce: bytes = secrets.token_bytes(32)
        mm: bytes = message + b"\x00" * (32 - len(message))
        G: bytes = XOR.repeated_key_xor(mm, hashlib.sha256(nonce).digest())
        H: bytes = XOR.repeated_key_xor(nonce, hashlib.sha256(G).digest())
        return G + H

    @staticmethod
    def oaep_unpad(message) -> bytes:
        """
        Un-pads a message
        :param message: The padded message.
        :return: The original message.
        """
        oaep_step2: bytes = message + b'\x00' * (64 - len(message))
        G: bytes = oaep_step2[:32]
        H: bytes = oaep_step2[32:64]
        nonce = XOR.repeated_key_xor(H, hashlib.sha256(G).digest())[:32]
        mm = XOR.repeated_key_xor(G, hashlib.sha256(nonce).digest())
        return mm.rstrip(b'\x00')



from __future__ import annotations
from . import ECPoint, ECKey
from .EllipticCurveNISTP256 import EllipticCurveNISTP256
from AsymmetricEncryptions.General import XOR
from AsymmetricEncryptions.Protocols import KDF
import secrets

class ECIES:
    """Elliptic Curve Integrated Encryption Scheme.
    This is how encryption really works in ECC.
    """

    def __init__(self):
        pass



    @staticmethod
    def encrypt(msg: bytes, pub_key: ECPoint, encryption_function=XOR.repeated_key_xor) -> tuple[bytes, ECPoint]:
        """
        Encrypts a message with a public ECC key (As an ECPoint)
        :param msg: The message.
        :param pub_key: The public key as an ECPoint (If in ECKey you can just do .public_key)
        :param encryption_function: The symmetric encryption function used (XOR is used here if no changes are made)
        :return: The ciphertext as a tuple
        """
        r: int = secrets.randbelow(EllipticCurveNISTP256().p)
        G: ECPoint = EllipticCurveNISTP256().g()
        R: ECPoint = G * r
        S: ECPoint = pub_key * r
        symmetric_key: bytes = bytes(S)
        symmetric_key: bytes = KDF.derive_key(symmetric_key)
        encrypted: bytes = encryption_function(msg, symmetric_key)
        return encrypted, R

    @staticmethod
    def decrypt(encrypted_msg, private_key: ECKey, decryption_function=XOR.repeated_key_xor) -> bytes:
        """
        Decrypts ciphertext.
        :param encrypted_msg: The ciphertext.
        :param private_key: The private key.
        :param decryption_function: The symmetric decryption function
        :return: message
        """
        encrypted, R = encrypted_msg
        S: ECPoint = R * private_key.private_key
        symmetric_key: bytes = bytes(S)
        symmetric_key: bytes = KDF.derive_key(symmetric_key)
        msg: bytes = decryption_function(encrypted, symmetric_key)
        return msg

if __name__ == '__main__':
    keyPair = ECKey.new()
    msg = b"test"
    c = ECIES.encrypt(msg, keyPair.public_key)
    print(c)
    d = ECIES.decrypt(c, keyPair)
    print(d)
    assert d == msg



from __future__ import annotations
from AsymmetricEncryptions.PublicPrivateKey.ECC import ECKey, EllipticCurveNISTP256, ECPoint
from AsymmetricEncryptions.General import BytesAndInts, PrimeNumberGen
import hashlib
import secrets
class ECDSA:
    """Elliptic Curve Digital Signature Algorithm, doesn't work, use ECSchnorr instead"""
    def __init__(self, privKey: ECKey) -> None:
        raise NotImplementedError("Code doesn't work for some odd reason")
        self.privKey: ECKey = privKey


    def sign(self, msg: bytes) -> tuple[int, int]:
        params = EllipticCurveNISTP256()
        e: int = BytesAndInts.byte2Int(hashlib.sha256(msg).digest())
        k: int = BytesAndInts.byte2Int(hashlib.sha256(msg + str(self.privKey).encode()).digest()) % params.n
        curve_point: ECPoint = params.g() * k
        r: int = curve_point.x % params.n
        inv_k: int = pow(k, -1, params.n)
        s: int = (inv_k * (e + (r * self.privKey.private_key))) % params.n
        if s == 0: self.sign(msg)
        return r, s

    @staticmethod
    def verify(pubKey: ECPoint, signature: tuple[int, int], msg: bytes) -> bool:
        params = EllipticCurveNISTP256()
        r, s = signature
        if pubKey.is_inf(): return False
        if not 1 < r < params.n or not 1 < s < params.n: return False
        e: int = BytesAndInts.byte2Int(hashlib.sha256(msg).digest())
        w = pow(s, -1, params.n)
        u1: int = (e * w) % params.n
        u2: int = (r * w) % params.n
        point: ECPoint = (params.g() * u1) + (pubKey * u2)
        if point.is_inf(): return False
        return r == point.x % params.n

if __name__ == '__main__':
    PrimeNumberGen.generate()
    AliceKey = ECKey()
    A = AliceKey.public_key
    ecdsa = ECDSA(AliceKey)
    msg = b'test'
    signature = ecdsa.sign(msg)
    v = ECDSA.verify(A, signature, msg)
    print(v)




from setuptools import setup, find_packages

DESC = 'A simple asymmetric encryption module'
with open("README.md", 'r') as f:
    LONG_DESC = f.read()
setup(
    name="asymmetric-encryption",
    version="0.0.6",
    author="Roy Nisimov, jacebalaron (Daniel gaisenberg)",
    description=DESC,
    long_description_content_type="text/markdown",
    long_description=LONG_DESC,
    packages=find_packages(),
    url='https://github.com/RoyNisimov/AsymmetricEncryption',
    license='MIT',
    install_requires=[],
    keywords=['python', 'cipher', 'asymmetric encryptions', 'signing', 'verifying', 'protocols', 'encryption', 'decryption'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



__all__ = ["PrimeNumberGen", "BytesAndInts", "XOR", "Exportation"]
from .PrimeNumberGen import PrimeNumberGen
from .BytesAndInts import BytesAndInts
from .XOR import XOR
from .Exportation import Exportation




from AsymmetricEncryptions.PublicPrivateKey.ECC import ECPoint, ECKey

class ECDH:
    """Elliptic Curve Diffie-Hellman"""
    def __init__(self, key_pairA: ECKey) -> None:
        """
        :param key_pairA: Alice's key pair
        """
        self.key_pairA: ECKey = key_pairA

    def Stage1(self, B: ECPoint) -> ECPoint:
        """
        :param B: Bob's public key
        :return: Shared ECPoint (Use KDF.derive_key(bytes(shared_point)) to get a key)
        """
        return B * self.key_pairA.private_key

    @staticmethod
    def Stage2(key_pairB: ECKey, A: ECPoint) -> ECPoint:
        """
        :param key_pairB: Bob's private key pair
        :param A: Alice's public key
        :return: Shared ECPoint (Use KDF.derive_key(bytes(shared_point)) to get a key)
        """
        return A * key_pairB.private_key


if __name__ == '__main__':
    keyA = ECKey.new()
    ecdh = ECDH(keyA)
    A = keyA.public_key

    keyB = ECKey()
    B = keyB.public_key

    shared_key_alice = ecdh.Stage1(B)

    shared_key_bob = ECDH.Stage2(keyB, A)

    print(shared_key_alice)
    print(shared_key_bob)

    assert shared_key_alice == shared_key_bob



import secrets
from AsymmetricEncryptions.General import BytesAndInts

class SSS:
    """
    Shamir's Secret Sharing over a finite field of 10**5.
    """

    FIELD_SIZE: int = pow(10, 5)

    @staticmethod
    def reconstruct_secret(shares: list[tuple[int, int]]) -> bytes:
        """
        Combines individual shares (points on graph)
        using Lagrange's interpolation.
        :param shares: `shares` is a list of points (x, y) belonging to a
        polynomial with a constant of our key. PLEASE ENTER ONLY T NUMBER OF POINTS OR THIS WON'T WORK!
        :return: The secret, assuming t points have been entered.
        """
        sums: int = 0
        for j, share_j in enumerate(shares):
            xj, yj = share_j
            prod: int = 1
            for i, share_i in enumerate(shares):
                xi, _ = share_i
                if i != j:
                    prod *= xi / (xi - xj)
            prod *= yj
            sums += prod
        a: int = int(round(sums, 0))
        return BytesAndInts.int2Byte(a)

    @staticmethod
    def polynom(x: int, coefficients: list[int]) -> int:
        """
        This generates a single point on the graph of given polynomial
        in `x`. The polynomial is given by the list of `coefficients`.
        """
        point: int = 0
        # Loop through reversed list, so that indices from enumerate match the
        # actual coefficient indices
        for coefficient_index, coefficient_value in enumerate(coefficients[::-1]):
            point += x ** coefficient_index * coefficient_value
        return point

    @staticmethod
    def coeff(t: int, secret: int) -> list[int]:
        """
        Randomly generate a list of coefficients for a polynomial with
        degree of `t` - 1, whose constant is `secret`.

        For example with a 3rd degree coefficient like this:
            3x^3 + 4x^2 + 18x + 554

            554 is the secret, and the polynomial degree + 1 is
            how many points are needed to recover this secret.
            (in this case it's 4 points).
        """
        coeff: list[int] = [secrets.SystemRandom().randrange(0, SSS.FIELD_SIZE) for _ in range(t - 1)]
        coeff.append(secret)
        return coeff

    @staticmethod
    def generate_shares(n: int, t: int, secret: bytes) -> list[tuple[int, int]]:
        """
        :param n: Number of shares to be split
        :param t: The number of required shares to reconstruct the secret.
        :param secret: The secret
        :return: List of cords (x, y).
        """
        secret: int = BytesAndInts.byte2Int(secret)
        coefficients: list[int] = SSS.coeff(t, secret)
        shares: list[tuple[int, int]] = []

        for i in range(1, n + 1):
            x = secrets.SystemRandom().randrange(1, SSS.FIELD_SIZE)
            shares.append((x, SSS.polynom(x, coefficients)))
        return shares



from __future__ import annotations
from AsymmetricEncryptions.General import PrimeNumberGen, Exportation, XOR
import secrets
from hashlib import sha256

class DLIESKey:

    def __init__(self, g: int, n: int, y: int, x: int = None):
        self.g: int = g
        self.n: int = n
        self.y: int = y
        self.x: int = x
        self.has_private: bool = False
        if x:
            self.has_private = True
            self.public: DLIESKey = DLIESKey(g, n, y)

    @staticmethod
    def new(nBit: int) -> DLIESKey:
        n: int = PrimeNumberGen.generate(nBit)
        g: int = secrets.randbelow(n)
        x: int = secrets.randbelow(n)
        y: int = pow(g, x, n)
        return DLIESKey(g, n, y, x)

    def export(self, file_name: str, pwd: bytes = b"\x00", *, enc_func=XOR.repeated_key_xor) -> None:
        data_dict: dict = {"g": self.g, "n": self.n, "x": self.x, "y": self.y}
        Exportation.export(file_name=file_name, pwd=pwd, data_dict=data_dict, exportation_func=enc_func)

    @staticmethod
    def load(file_name: str, pwd: bytes = b"\x00", *, dec_func=XOR.repeated_key_xor) -> DLIESKey:
        return DLIESKey(**Exportation.load(file_name=file_name, pwd=pwd, dec_func=dec_func))

    def __eq__(self, other: DLIESKey) -> bool:
        if not isinstance(other, DLIESKey): return False
        return sha256(f"{self}".encode()).hexdigest() == sha256(f"{other}".encode()).hexdigest()

    def __str__(self) -> str:
        r: str = ""
        if self.has_private:
            r += f"""
    Private Key:

    g = {self.g}
    n = {self.n}
    x = {self.x}
    y = {self.y}



    """
            r += f"""
    Public Key:

    g = {self.g}
    n = {self.n}
    y = {self.y}

    """
            return r


from __future__ import annotations
from AsymmetricEncryptions.General import PrimeNumberGen, BytesAndInts
from AsymmetricEncryptions.PublicPrivateKey.RSA import RSAKey, RSA
import secrets
class DiffieHellman:
    """RSA Diffie-Hellman, (Discrete Logarithms Diffie Hellman).
    Example:
    ```
    Apriv, Apub = RSA.generate_key_pair(2048)
    Bpriv, Bpub = RSA.generate_key_pair(2048)
    DH = DiffieHellman.new(Apriv, 2048)
    # Alice
    A = DH.Stage1()
    gp = DH.get_gp()
    # send A and gp to Bob

    # Bob
    B, Bob_shared = DiffieHellman.Stage2(gp, Bpriv, A)
    print(Bob_shared)
    # send B to Alice

    # Alice
    Alice_shared = DH.Stage3(B)
    print(Alice_shared)
    print(Alice_shared == Bob_shared)
    ```

    """
    def __init__(self, priv_key: RSAKey, p: int, g: int) -> None:
        self.key: RSAKey = priv_key
        self.p: int = p
        self.g: int = g

    def get_gp(self) -> tuple[int, int]:
        return self.g, self.p

    @staticmethod
    def new(priv_key: RSAKey, nBitP: int) -> DiffieHellman:
        """
        :param priv_key: The first's person private key.
        :param nBitP: How big p is.
        :return: DiffieHellman Object.
        """
        p: int = PrimeNumberGen.generate(nBitP)
        g: int = secrets.randbelow(p)
        return DiffieHellman(priv_key, p, g)

    def Stage1(self) -> int:
        return pow(self.g, self.key.d, self.p)

    @staticmethod
    def Stage2(gp: tuple[int, int], Key: RSAKey, A: int) -> int and bytes:
        shared_key: int = pow(A, Key.d, gp[1])
        B: int = pow(gp[0], Key.d, gp[1])
        return B, BytesAndInts.int2Byte(shared_key)

    def Stage3(self, B) -> bytes:
        shared_key: int = pow(B, self.key.d, self.p)
        return BytesAndInts.int2Byte(shared_key)
        # shared key is g**(a*b) % p

if __name__ == '__main__':
    Apriv, Apub = RSA.generate_key_pair(2048)
    Bpriv, Bpub = RSA.generate_key_pair(2048)
    DH = DiffieHellman.new(Apriv, 2048)
    # Alice
    A = DH.Stage1()
    gp = DH.get_gp()
    # send A and gp to Bob

    # Bob
    B, Bob_shared = DiffieHellman.Stage2(gp, Bpriv, A)
    print(Bob_shared)
    # send B to Alice

    # Alice
    Alice_shared = DH.Stage3(B)
    print(Alice_shared)
    print(Alice_shared == Bob_shared)






from __future__ import annotations
from AsymmetricEncryptions.PublicPrivateKey.RSA.RSAKey import RSAKey
from AsymmetricEncryptions.General import BytesAndInts

class RSA:
    """Rivest-Shamir-Adleman"""
    def __init__(self, key: RSAKey) -> None:
        self.key = key

    @staticmethod
    def generate_key_pair(nBit) -> RSAKey and RSAKey:
        """
        RSAKey.new returns a pair of priv and pub
        :param nBit: Key length (2048+ is recommended )
        :return: Private key, public key
        """
        Priv: RSAKey = RSAKey.new(nBit)
        Pub: RSAKey = Priv.public
        return Priv, Pub


    def encrypt(self, msg: bytes, assertion=True) -> bytes:
        """
        Encrypts a message
        :param msg: The plaintext
        :param assertion: If assertion is true then it will assert if the message is smaller than allowed. (e.g. m < p)
        :return: Ciphertext
        """
        int_msg: int = BytesAndInts.byte2Int(msg)
        if assertion: assert int_msg < self.key.n
        cipher: int = pow(int_msg, self.key.e, self.key.n)
        return BytesAndInts.int2Byte(cipher)

    def decrypt(self, cipher: bytes) -> bytes:
        """
        Decrypts (Only possible if key is private)
        :param cipher: the ciphertext
        :return: message
        """
        assert self.key.has_private
        int_cipher: int = BytesAndInts.byte2Int(cipher)
        m: int = pow(int_cipher, self.key.d, self.key.n)
        return BytesAndInts.int2Byte(m)

    def sign(self, msg: bytes, assertion=True) -> bytes:
        """
        Signs a message
        :param msg: The plaintext
        :param assertion: if true then will check if m is allowed
        :return: signature
        """
        assert self.key.has_private
        int_msg: int = BytesAndInts.byte2Int(msg)
        if assertion: assert int_msg < self.key.n
        s: int = pow(int_msg, self.key.d, self.key.n)
        return BytesAndInts.int2Byte(s)

    def verify(self, signature: bytes, og_msg: bytes) -> None:
        """
        Verifies a signature.
        :param signature: The signature.
        :param og_msg: The original message.
        :return: None, will throw an assertion error if fails
        """
        int_s: int = BytesAndInts.byte2Int(signature)
        cipher: int = pow(int_s, self.key.e, self.key.n)
        assert cipher == BytesAndInts.byte2Int(og_msg)



from AsymmetricEncryptions.Protocols.KDF import KDF
from AsymmetricEncryptions.General import XOR
import json
from secrets import compare_digest, token_bytes
import hmac
import base64
from hashlib import sha256
class Exportation:
    """
    Handles the exportation of keys. Nothing in here is done by regulations though.
    """


    @staticmethod
    def export(file_name: str, data_dict: dict, pwd: bytes, *, exportation_func=XOR.repeated_key_xor) -> None:
        """
        Encrypts and exports data that can be converted to json.
        :param file_name: The file that will be exported into
        :param data_dict: The data that will be exported
        :param pwd: The passphrase that will be used in the KDF and then encryption part of the export
        :param exportation_func: The encryption function used (I'm using XOR or OTP)
        :return: None
        """
        key: bytes = KDF.derive_key(pwd)
        jData: bytes = json.dumps(data_dict).encode("utf-8")
        randomness: bytes = sha256(token_bytes(32)).digest()
        xored: bytes = XOR.repeated_key_xor(jData, randomness)
        write_data: bytes = exportation_func(xored, key)
        mac: hmac = hmac.new(key=key, msg=jData, digestmod="sha512")
        final_data: bytes = mac.digest() + write_data + randomness
        final_data = base64.b64encode(final_data)
        with open(file_name, "wb") as f:
            f.write(final_data)

    @staticmethod
    def load(file_name: str, pwd: bytes, *, dec_func=XOR.repeated_key_xor) -> dict:
        """
        Loads data from an encrypted file
        :param file_name: The encrypted file name
        :param pwd: The passphrase used as a key
        :param dec_func: The opposite of the encryption function
        :return: The data as a dict
        """
        key: bytes = KDF.derive_key(pwd)
        with open(file_name, "rb") as f:
            final_data: bytes = f.read()
        final_data = base64.b64decode(final_data)
        dMac: bytes = final_data[:64]
        read_data: bytes = final_data[64:-32]
        randomness: bytes = final_data[-32:]
        jData: bytes = dec_func(read_data, key)
        jData = XOR.repeated_key_xor(jData, randomness)
        mac: hmac = hmac.new(key=key, msg=jData, digestmod="sha512")
        assert compare_digest(mac.digest(), dMac)
        return json.loads(jData.decode())



from AsymmetricEncryptions.General import XOR, BytesAndInts
from AsymmetricEncryptions.Protocols import KDF
from .DLIESKey import DLIESKey
import secrets

class DLIES:
    """
    Discrete Logarithm Integrated Encryption Scheme.
    """

    def __init__(self) -> None:
        pass

    @staticmethod
    def generate_key_pair(nBits: int) -> tuple[DLIESKey, DLIESKey]:
        """
        Generates a DLIES key pair
        :param nBits: How large the modulus be.
        :return: Private key, Public Key
        """
        priv: DLIESKey = DLIESKey.new(nBits)
        return priv, priv.public

    @staticmethod
    def encrypt(pub_key: DLIESKey, msg: bytes, encryption_func=XOR.repeated_key_xor) -> tuple[bytes, int]:
        """
        Encrypts the message.
        :param pub_key: Public key
        :param msg: The plaintext
        :param encryption_func: The symmetric encryption function
        :return: Encrypted Message, Decrypt helper as a tuple, don't split.
        """
        r: int = secrets.randbelow(pub_key.n)
        R: int = pow(pub_key.g, r, pub_key.n)
        S: int = pow(pub_key.y, r, pub_key.n)
        S_bytes: bytes = BytesAndInts.int2Byte(S)
        key: bytes = KDF.derive_key(S_bytes)
        encrypted = encryption_func(msg, key)
        return encrypted, R

    @staticmethod
    def decrypt(private_key: DLIESKey, ciphertxt: tuple[bytes, int], decryption_function=XOR.repeated_key_xor) -> bytes:
        """
        Decrypts a ciphertext.
        :param private_key: The private key.
        :param ciphertxt: The ciphertext containing (encrypted, R)
        :param decryption_function: The symmetric decryption function
        :return: The message
        """
        encrypted, R = ciphertxt
        assert private_key.x and private_key.has_private
        S: int = pow(R, private_key.x, private_key.n)
        S_bytes: bytes = BytesAndInts.int2Byte(S)
        key: bytes = KDF.derive_key(S_bytes)
        return decryption_function(encrypted, key)
